#!/usr/bin/env python3

class ConfigurationStrings():
    DOMAIN = "Domain"
    USERNAME = "Username"
    PASSWORD = "Password"
    TASKS = "Tasks"
    SEARCH_OBJECT = "search_object"
    SEARCH_STRING = "search_string"
    COURSES = "courses"
    ILIAS = "ilias"
    TYPE = "type"



from typing import Any

from pydantic import BaseModel
from datetime import datetime

class Result(BaseModel):
    failed: bool

class ItemTitle(BaseModel):
    text: str
    href: str

class ItemLocation(BaseModel):
    pass

class ItemAllLocation(ItemLocation):
    pass

class ItemTitleLocation(ItemLocation):
    pass

class ItemTitleTextLocation(ItemTitleLocation):
    pass

class ItemTitleHrefLocation(ItemTitleLocation):
    pass

class ItemDescriptionLocation(ItemLocation):
    pass

class ItemPropertyLenLocation(ItemLocation):
    pass

class ItemPropertyLocation(ItemLocation):
    def __init__(self, index):
        self.index = index

class IliasListItem(BaseModel):
    title: ItemTitle
    description: str
    properties: list[str]

    def __init__(self, list_item: Any) -> None:
        t = list_item.find('a', class_='il_ContainerItemTitle')
        self.title = ItemTitle(text=t.text, href=t.href)
        self.description = list_item.find('a', class_='il_Description')
        self.properties = [item_property.text for item_property in list_item.find_all('span', class_='il_ItemProperty')]

    def match_string(self, string: str) -> list[ItemLocation]:
        locations: list[ItemLocation] = []
        if string in self.title.text:
            locations.append(ItemTitleTextLocation())
        if string in self.title.href:
            locations.append(ItemTitleHrefLocation())
        if string in self.description:
            locations.append(ItemDescriptionLocation())
        for i, prop in enumerate(self.properties):
            if string in prop:
                locations.append(ItemPropertyLocation(i))
        return locations

class DifferenceEvent(BaseModel):
    locations: list[ItemLocation]

class AddEvent(DifferenceEvent):
    pass

class DeleteEvent(DifferenceEvent):
    pass

class ChangeEvent(DifferenceEvent):
    pass

class IliasListItemDifference(BaseModel):
    before: IliasListItem | None
    after: IliasListItem | None

    def __init__(self, before: IliasListItem | None, after: IliasListItem | None) -> None:
        self.before = before
        self.after = after

    def get_changes(self) -> list[DifferenceEvent]:
        if self.is_equal():
            return []

        if self.before is None and self.after is not None:
            return [AddEvent(locations=[ItemAllLocation()])]

        if self.before is not None and self.after is None:
            return [DeleteEvent(locations=[ItemAllLocation()])]

        # This should always be true because of the is_equal call at the start
        assert(self.before is not None and self.after is not None)

        change_locations: list[ItemLocation] = []

        if self.before.title.text != self.after.title.text:
            change_locations.append(ItemTitleTextLocation())
        if self.before.title.href != self.after.title.href:
            change_locations.append(ItemTitleHrefLocation())
        if self.before.description != self.after.description:
            change_locations.append(ItemDescriptionLocation())

        before_prop_n = len(self.before.properties)
        after_prop_n = len(self.after.properties)
        if before_prop_n != after_prop_n:
            change_locations.append(ItemPropertyLenLocation())
            return [ChangeEvent(locations=change_locations)]

        for i in range(before_prop_n):
            if self.before.properties[i] != self.after.properties[i]:
                change_locations.append(ItemPropertyLocation(index=i))
        return [ChangeEvent(locations=change_locations)]
        

    def match_string(self, string: str) -> list[DifferenceEvent]:
        if self.is_equal():
            return []

        if self.before is None and self.after is not None:
            add_locations = self.after.match_string(string)
            if len(add_locations) != 0:
                return [AddEvent(locations=add_locations)]
            return []

        if self.before is not None and self.after is None:
            delete_locations = self.before.match_string(string)
            if len(delete_locations) != 0:
                return [DeleteEvent(locations=delete_locations)]
            return []

        # This should always be true because of the is_equal call at the start
        assert(self.before is not None and self.after is not None)

        locations_before = self.before.match_string(string)
        locations_after = self.after.match_string(string)

        all_locations: set[ItemLocation] = set()
        all_locations.update(locations_before)
        all_locations.update(locations_after)

        change_locations: list[ItemLocation] = []
        add_locations: list[ItemLocation] = []
        delete_locations: list[ItemLocation] = []
        for location in all_locations:
            if location in locations_before and location in locations_after:
                change_locations.append(location)
                continue
            if location in locations_before:
                delete_locations.append(location)
            if location in locations_after:
                add_locations.append(location)

        events: list[DifferenceEvent] = []
        if len(change_locations) != 0:
            events.append(ChangeEvent(locations=change_locations))
        if len(add_locations) != 0:
            events.append(AddEvent(locations=add_locations))
        if len(delete_locations) != 0:
            events.append(DeleteEvent(locations=delete_locations))
        return events

    def is_equal(self) -> bool:
        return self.before == self.after

class CourseSnapshot(BaseModel):
    object_id: str
    title: str
    items: list[IliasListItem]
    time: datetime

class CourseDiff(BaseModel):
    title_change: tuple[str | None, str | None]
    itemchanges_by_title: dict[str, IliasListItemDifference]

    def __init__(self, before: CourseSnapshot | None, after: CourseSnapshot | None) -> None:
        def items_by_title(items: list[IliasListItem]) -> dict[str, IliasListItem]:
            return {item.title.text: item for item in items}

        changes: dict[str, IliasListItemDifference] = {}
        before_by_items = items_by_title(before.items) if before else {}
        after_by_items = items_by_title(after.items) if after else {}

        all_keys: set[str] = set()
        all_keys.update(before_by_items.keys())
        all_keys.update(after_by_items.keys())

        for title in all_keys:
            item_before = before_by_items.get(title)
            item_after = after_by_items.get(title)
            diff = IliasListItemDifference(before=item_before, after=item_after)
            if not diff.is_equal():
                changes[title] = diff

        self.itemchanges_by_title = changes
        title_change: list[str | None] = [None, None]
        if before:
            title_change[0] = before.title
        if after:
            title_change[1] = after.title
        self.title_change = tuple(title_change)

    def search_matching_listitems(self, search_string: str) -> dict[str, list[DifferenceEvent]]:
        matching_listitems: dict[str, list[DifferenceEvent]] = {}
        for title, change in self.itemchanges_by_title.items():
            match: list[DifferenceEvent] = change.match_string(search_string)
            if len(match):
                matching_listitems[title] = match
        return matching_listitems

class UpdateResult(Result):
    diff: CourseDiff

class AvailabilityResult(Result):
    availables: dict[str, list[DifferenceEvent]]

class CourseResult(BaseModel):
    update: UpdateResult | None
    availability: list[AvailabilityResult]

class WatchResult(BaseModel):
    courses: dict[str, CourseResult]




from time import sleep
from bs4 import BeautifulSoup, Tag
import requests
from pydantic import BaseModel
from datetime import datetime
import bisect
import logging

from .Configuration import WatchConfig, IliasConfig, AvailabilityTask, UpdateTask
from .Result import CourseSnapshot, CourseResult, IliasListItem, IliasListItemDifference, UpdateResult, AvailabilityResult, CourseDiff, WatchResult

logger = logging.getLogger(__name__)

class Memory:

    def __init__(self):
        self._mem: dict[str, list[CourseSnapshot]] = {}

    def remember(self, snapshot: CourseSnapshot) -> None:
        bisect.insort(self._mem[snapshot.object_id], snapshot, key=lambda s: s.time)

    def get_latest(self, object_id: str) -> CourseSnapshot | None:
        try:
            return self._mem[object_id][-1]
        except KeyError or IndexError:
            return None

    def get_previous(self, object_id: str) -> CourseSnapshot | None:
        try:
            return self._mem[object_id][-2]
        except KeyError or IndexError:
            return None


class ConnectionError(Exception):
    pass

class ParseError(Exception):
    pass

class CourseWatcher:

    def __init__(self, ilias: IliasConfig, watchconf: WatchConfig, memory: Memory) -> None:
        self.ilias = ilias
        self.watchconf = watchconf
        self.session = requests.Session()
        self.memory = memory
        self.session = requests.Session()

    def _is_logged_in(self) -> bool:
        resp = self.session.get(self.ilias.domain)
        if resp.status_code != 200:
            raise ConnectionError(f'Failed to connect to the domain {self.ilias.domain}')
        return b'href="https://www.studon.fau.de/studon/login.php' in resp.content

    def _login(self):
        if self._is_logged_in():
            return
        self.session.post(
            'https://www.studon.fau.de/studon/ilias.php?lang=de&cmd=post&cmdClass=ilstartupgui&cmdNode=16d&baseClass=ilStartUpGUI&rtoken=',
            data={
                'username': self.ilias.username,
                'password': self.ilias.password,
                'cmd[doStandardAuthentication]': 'Anmelden'
            }
        )
        if not self._is_logged_in():
            logger.error("Failed to log in.")
            raise ConnectionError('Failed to log in.')

    def watch(self) -> WatchResult:
        course_results: dict[str, CourseResult] = {}

        def get_html_txt(obj: str) -> str:
            url = self.ilias.domain + '/' + obj + '.html'
            n = 5
            for i in range(n):
                self._login()
                resp: requests.Response = self.session.get(url)
                if resp.status_code not in [200, 302]:
                    logger.warning(f'Could not fetch "{url}". Status code: {resp.status_code} | Body: {resp.content}')
                    sleep(5)
                    continue
                return resp.text
            raise ConnectionError(f'Failed to fetch {url} after multiple {n} tries')

        for object_id, crsconf in self.watchconf.courses.items():
            html_resp = get_html_txt(object_id)
            soup = BeautifulSoup(html_resp, 'html.parser')
            items = soup.find_all('div', class_='il_ContainerListItem')

            listitems_list: list[IliasListItem] = []
            for item in items:
                list_item = IliasListItem(item)
                listitems_list.append(list_item)

            title = soup.find('a', {'id': 'il_mhead_t_focus'})
            if not isinstance(title, Tag):
                raise ParseError(f"Title for course with object id {object_id} is not a tag.")

            current = CourseSnapshot(object_id=object_id, title=title.text, items=listitems_list, time=datetime.now())
            self.memory.remember(current)

            crs_result = CourseResult(update=None, availability=[])

            current = self.memory.get_latest(object_id)
            previous = self.memory.get_previous(object_id)
            crsdiff: CourseDiff = CourseDiff(
                after=current,
                before=previous
            )

            if crsconf.update_task:
                crs_result.update = UpdateResult(failed=False, diff=crsdiff)

            for availability_task in crsconf.availability_tasks:
                availability_result = AvailabilityResult(failed=False, availables=crsdiff.search_matching_listitems(availability_task.search_string))
                crs_result.availability.append(
                    availability_result
                )

            course_results[object_id] = crs_result
        return WatchResult(courses=course_results)



__version__ = "0.2.0"

from .CourseWatcher import CourseWatcher
from .Configuration import Stakeholders, UpdateTask, AvailabilityTask, CourseConfig, WatchConfig, IliasConfig
from .ConfigurationStrings import ConfigurationStrings



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='iliascw',
    version='0.2.0',
    packages=find_packages(),
    install_requires=[
        'beautifulsoup4==4.12.0',
        'certifi==2022.12.7',
        'charset-normalizer==3.1.0',
        'idna==3.4',
        'pydantic==1.10.7',
        'requests==2.28.2',
        'soupsieve==2.4',
        'typing_extensions==4.5.0',
        'urllib3==1.26.15',
    ],
    author='Merlin Sievers',
    author_email='merlin.sievers@posteo.net',
    description='Library to check if courses are available.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/dann-merlin/ilias-course-watcher',
    classifiers=[
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
)



#!/usr/bin/env python3

from abc import abstractmethod
import os
from pydantic import BaseModel
import logging
import json

from .ConfigurationStrings import ConfigurationStrings as CS
from .IliasConfig import ILIAS_USERNAME, ILIAS_PASSWORD, ILIAS_DOMAIN

logger = logging.getLogger('uvicorn')

class Stakeholders(BaseModel):
    telegramNotifications: set[str]
    mailNotifications: set[str]

    def include(self, other: 'Stakeholders'):
        self.telegramNotifications = self.telegramNotifications.union(other.telegramNotifications)
        self.mailNotifications = self.mailNotifications.union(other.mailNotifications)

    def clear(self):
        self.telegramNotifications.clear()
        self.mailNotifications.clear()

class Task(BaseModel):
    stakeholders: Stakeholders

    @abstractmethod
    def merge_into_course(self, course: 'CourseConfig') -> None:
        raise NotImplementedError("virtual function")

class UpdateTask(Task):

    def merge_into_course(self, course: 'CourseConfig') -> None:
        if course.update_task is None:
            logger.debug("Matching update task was not found. Replacing with constructed update task.")
            course.update_task = self
        else:
            logger.debug("Matching update task was found. Including the new stakeholders in the current task.")
            course.update_task.stakeholders.include(self.stakeholders)

class AvailabilityTask(Task):
    search_string: str

    def merge_into_course(self, course: 'CourseConfig') -> None:
        found_course = None
        for availability_task in course.availability_tasks:
            if availability_task.search_string == self.search_string:
                found_course = availability_task
                break

        if found_course is None:
            logger.debug("Matching availability task was not found. Replacing with constructed availability task.")
            course.availability_tasks.append(self)
        else:
            logger.debug("Matching availability task was found. Including the new stakeholders in the current task.")
            found_course.stakeholders.include(self.stakeholders)

class CourseConfig(BaseModel):
    update_task: UpdateTask | None
    availability_tasks: list[AvailabilityTask]

class WatchConfig(BaseModel):
    courses: dict[str, CourseConfig]

    def register_task(self, object_id: str, task: Task) -> None:
        if object_id not in self.courses.keys():
            self.courses[object_id] = CourseConfig(
                update_task=None,
                availability_tasks=[]
            )

        task.merge_into_course(self.courses[object_id])

class IliasConfig(BaseModel):
    username: str
    password: str
    domain: str

